import numpy as np
from skimage.morphology import white_tophat as skimage_white_tophat
from skimage.exposure import rescale_intensity
from skimage.restoration import richardson_lucy


def white_tophat(image, radius):
    """
    """

    selem = np.ones((radius,)*image.ndim)
    return skimage_white_tophat(image, selem=selem)


def rl_decon(image, psf, **kwargs):
    """
    """

    norm_image = rescale_intensity(
        image,
        in_range=(0, image.max()),
        out_range=(0, 1),
    )
    return richardson_lucy(norm_image, psf, **kwargs)




import numpy as np
from scipy.ndimage import maximum_filter


def get_context(spot, image, radius):
    """
    """

    p, r = spot, radius
    w = image[
        p[0]-r:p[0]+r+1,
        p[1]-r:p[1]+r+1,
        p[2]-r:p[2]+r+1
    ]
    if np.product(w.shape) != (2*r+1)**3:
        return None
    return w.flatten()


def estimate_psf(image, radius, n_spots=2000, **kwargs):
    """
    """

    # get weak spots
    max_image = maximum_filter(image, size=radius)
    spots = np.column_stack(np.nonzero(image == max_image))
    intensities = image[spots[:, 0], spots[:, 1], spots[:, 2]]
    sort_idxs = np.argsort(intensities)[::-1]
    spots = spots[sort_idxs[:n_spots], :]

    # get context around each weak spot
    blocks = []
    for spot in spots:
        context = get_context(spot, image, radius)
        if context is not None:
            blocks.append(context)
    blocks = np.array(blocks)

    # set ransac filter defaults
    if 'n_propose_model' not in kwargs:
        kwargs['n_propose_model'] = 10
    if 'max_iterations' not in kwargs:
        kwargs['max_iterations'] = 1000
    if 'inlier_threshold' not in kwargs:
        kwargs['inlier_threshold'] = 0.9
    if 'min_inliers' not in kwargs:
        kwargs['min_inliers'] = 25

    # estimate psf with ransac filter
    psf, inliers = ransac(blocks, model=psf_model, **kwargs)
    psf = psf.get_array(radius)

    # return normalized psf
    return psf / np.sum(psf)


class psf_model:

    def __init__(self):
        None

    def _norm_range(self, data):
        mx = np.max(data, axis=1)[..., None]
        mn = np.min(data, axis=1)[..., None]
        return (data - mn) / (mx - mn)

    def _stats(self, data):
        mean = np.mean(data, axis=1, dtype=np.float64)
        std = np.std(data, axis=1, dtype=np.float64)
        return mean, std

    def fit(self, data):
        norm_data = self._norm_range(data)
        self.psf = np.mean(norm_data, axis=0)
        return self

    def get_error(self, data):
        psf_mean, psf_std = self._stats(self.psf[None, ...])
        psf_cent = self.psf - psf_mean
        norm_data = self._norm_range(data)
        dat_mean, dat_std = self._stats(norm_data)
        dat_cent = norm_data - dat_mean[..., None]
        corr = -np.mean(dat_cent*psf_cent, axis=1)/dat_std/psf_std
        return corr.squeeze()

    def get_array(self, radius):
        return np.reshape(self.psf, (2*radius+1,)*3)


def ransac(
    data,
    n_propose_model,
    max_iterations,
    inlier_threshold,
    min_inliers,
    model=psf_model,
    verbose=False,
):
    """
    """

    # initialize
    iterations = 0
    bestfit = None
    besterr = np.inf
    best_inlier_idxs = None
    psf = model()

    # ensure inlier_threshold is a negative correlation
    if 0 < inlier_threshold < 1:
        inlier_threshold *= -1

    # loop over model proposals
    while iterations < max_iterations:

        # randomly partition data
        all_idxs = np.arange(data.shape[0])
        np.random.shuffle(all_idxs)
        maybe_idxs = all_idxs[:n_propose_model]
        test_idxs = all_idxs[n_propose_model:]
        maybeinliers = data[maybe_idxs]
        test_points = data[test_idxs]

        # fit model, get error
        maybemodel = psf.fit(maybeinliers)
        test_err = psf.get_error(test_points)

        # get fellow inliers
        also_idxs = test_idxs[test_err < inlier_threshold]
        alsoinliers = data[also_idxs]

        # provide feedback if requested
        if verbose:
            print('test_err.min()',test_err.min())
            print('test_err.max()',test_err.max())
            print('np.mean(test_err)',np.mean(test_err))
            print('iteration %d:len(alsoinliers) = %d'%(
                iterations,len(alsoinliers)))

        # update model with new inliers
        if len(alsoinliers) > min_inliers:
            betterdata = np.concatenate(
                (maybeinliers, alsoinliers)
            )
            bettermodel = psf.fit(betterdata)
            better_errs = psf.get_error(betterdata)

            # get updated model error, if best overwrite old model
            thiserr = np.mean( better_errs )
            if thiserr < besterr:
                bestfit = bettermodel
                besterr = thiserr
                best_inlier_idxs = np.concatenate( (maybe_idxs, also_idxs) )

        # tick tock
        iterations+=1

    # final results
    if bestfit is None:
        raise ValueError("did not meet fit acceptance criteria")
    return bestfit, {'inliers':best_inlier_idxs}





import numpy as np
from skimage.feature import blob_log


def detect_spots_log(
    image,
    min_blob_radius,
    max_blob_radius,
    **kwargs,
):
    """
    """

    # compute defaults
    min_blob_radius = np.array(min_blob_radius)
    max_blob_radius = np.array(max_blob_radius)
    min_sigma = 0.8 * min_blob_radius / np.sqrt(3)
    max_sigma = 1.2 * max_blob_radius / np.sqrt(3)
    num_sigma = int(np.ceil(np.max(max_blob_radius - min_blob_radius)))

    # set defaults
    if 'min_sigma' not in kwargs:
        kwargs['min_sigma'] = min_sigma
    if 'max_sigma' not in kwargs:
        kwargs['max_sigma'] = max_sigma
    if 'num_sigma' not in kwargs:
        kwargs['num_sigma'] = num_sigma
    if 'overlap' not in kwargs:
        kwargs['overlap'] = 0.8

    # run
    return blob_log(image, **kwargs)




import setuptools

setuptools.setup(
    name="fishspot",
    version="0.2.1",
    author="Greg M. Fleishman",
    author_email="greg.nli10me@gmail.com",
    description="Tools for finding discrete bright spots in images",
    url="https://github.com/GFleishman/fishspot",
    license="MIT",
    packages=setuptools.find_packages(),
    include_package_data=True,
    install_requires=[
        'numpy',
        'scipy',
        'scikit-image',
        'ClusterWrap',
        'zarr',
        'dask',
        'dask[array]',
    ]
)



import numpy as np
import dask
import dask.array as da
import dask.delayed as delayed
import ClusterWrap
import time
import fishspot.filter as fs_filter
import fishspot.psf as fs_psf
import fishspot.detect as fs_detect
import fishspot.assign as fs_assign


def distributed_spot_detection(
    array, blocksize,
    white_tophat_args={},
    psf_estimation_args={},
    deconvolution_args={},
    spot_detection_args={},
    cluster_kwargs={},
    mask=None,
    psf=None,
    psf_retries=3,
):
    """
    """

    # set white_tophat defaults
    if 'radius' not in white_tophat_args.keys():
        white_tophat_args['radius'] = 9

    # set psf estimation defaults
    if 'radius' not in psf_estimation_args.keys():
        psf_estimation_args['radius'] = 9

    # set deconvolution defaults
    if 'filter_epsilon' not in deconvolution_args.keys():
        deconvolution_args['filter_epsilon'] = 1e-8

    # set spot detection defaults
    if 'min_blob_radius' not in spot_detection_args.keys():
        spot_detection_args['min_blob_radius'] = 3
    if 'max_blob_radius' not in spot_detection_args.keys():
        spot_detection_args['max_blob_radius'] = 6

    # compute overlap depth
    overlap = 2*spot_detection_args['max_blob_radius']

    # compute mask to array ratio
    if mask is not None:
        ratio = np.array(mask.shape) / array.shape

    # pipeline to run on each block
    def detect_spots_pipeline(block, mask=None, psf=None, block_info=None):

        # get origin (used a few times later)
        origin = np.array(block_info[0]['chunk-location'])
        origin = origin * blocksize - overlap

        # check mask
        if mask is not None:
            mo = np.round(origin * ratio).astype(np.uint16)
            mo = np.maximum(0, mo)
            ms = np.round(blocksize * ratio).astype(np.uint16)
            mask_block = mask[mo[0]:mo[0]+ms[0],
                              mo[1]:mo[1]+ms[1],
                              mo[2]:mo[2]+ms[2],]

            # if there is no foreground, return null result
            if np.sum(mask_block) < 1:
                result = np.empty((1,1,1), dtype=list)
                result[0, 0, 0] = [np.zeros((0, 4)), psf]
                return result

        # background subtraction: white tophat filter
        wth = fs_filter.white_tophat(block, **white_tophat_args)

        if psf is None:
            # automated psf estimation with error handling
            for i in range(psf_retries):
                try:
                    psf = fs_psf.estimate_psf(wth, **psf_estimation_args)
                except ValueError:
                    if 'inlier_threshold' not in psf_estimation_args.keys():
                        psf_estimation_args['inlier_threshold'] = 0.9
                    psf_estimation_args['inlier_threshold'] -= 0.1
                else: break

        # deconvolution
        decon = fs_filter.rl_decon(wth, psf, **deconvolution_args)

        # blob detection
        spots = fs_detect.detect_spots_log(decon, **spot_detection_args)

        # make sure dimensions are compatible
        if spots.shape == (0, 3):
            spots = np.zeros((0, 4))

        # remove spots in the halo
        for i in range(3):
            spots = spots[spots[:, i] > overlap - 1]
            spots = spots[spots[:, i] < overlap + blocksize[i]]

        # adjust for block origin
        if spots.shape[0] > 0:
            spots[:, :3] = spots[:, :3] + origin

        # package and return
        result = np.empty((1,1,1), dtype=list)
        result[0, 0, 0] = [spots, psf]
        return result

    # construct cluster
    with ClusterWrap.cluster(**cluster_kwargs) as cluster:

        # a test dataset as a numpy array
        if isinstance(array, np.ndarray):
            future = cluster.client.scatter(array)
            array_da = da.from_delayed(
                future, shape=array.shape, dtype=array.dtype,
            )
            array_da = array_da.rechunk(blocksize)
            array_da.persist()
            time.sleep(5)  ### a little time for workers to be allocated
            cluster.client.rebalance()

        # a full dataset as a zarr array
        else:
            array_da = da.from_array(array, chunks=blocksize)

        # delay mask
        mask_d = delayed(mask) if mask is not None else None

        # execute spot detection on all blocks
        spots_as_grid = da.map_overlap(
            detect_spots_pipeline, array_da,
            mask=mask_d,
            psf=psf,
            depth=overlap,
            dtype=list,  ### dask generalizes to object
            boundary=0,
            trim=False,
            chunks=(1,1,1),
        ).compute()

    # reformat spots and psf into lists
    spots_list, psf_list = [], []
    for ijk in range(np.prod(spots_as_grid.shape)):
        i, j, k = np.unravel_index(ijk, spots_as_grid.shape)
        spots_list.append(spots_as_grid[i, j, k][0])
        psf_list.append(spots_as_grid[i, j, k][1])

    # concatenate all spots into single array
    spots = np.vstack(spots_list)

    # filter with foreground mask
    if mask is not None:
        spots = fs_assign.apply_foreground_mask(
            spots, mask, ratio,
        )

    # average over all psf estimates
    psf = np.mean(psf_list, axis=0)

    # return results
    return spots, psf




import numpy as np


def apply_foreground_mask(spots, mask, ratio):
    """
    """

    # get spot locations in mask voxel coordinates
    x = np.round(spots[:, :3] * ratio).astype(np.uint16)

    # correct out of range rounding errors
    for i in range(3):
        x[x[:, i] >= mask.shape[i], i] = mask.shape[i] - 1

    # filter spots and return
    return spots[mask[x[:, 0], x[:, 1], x[:, 2]] == 1]


def filter_by_range(spots, origin, span):
    """
    """

    # operate on a copy, filter lower/upper range all axes
    result = np.copy(spots)
    for i in range(3):
        result = result[result[:, i] > origin[i] - 1]
        result = result[result[:, i] < origin[i] + span[i]]
    return result


